
import axiomic.frontend.text as text


def format(template: str, **kwargs) -> text.Text:
    '''
    See `axiomic.format` for more information.

    .. code-block:: python

        name = axiomic.Text('World')
        assert 'Hello, World!' == format('Hello, {name}!', name=name).value()

    '''
    return text.Text(template).format(**kwargs)

def infer(prompt: any, system_prompt=None, history_pairs=[], name=None) -> text.Text:
    '''
    Does text completion on the given prompt.

    The model used for completion is determined by the context, for example,

    .. code-block:: python

        import axiomic.quick_config as qc
        with qc.OpenAiGpt4:
            assert 'Pairs' in infer('What is the capital of France?').value()

    .. code-block:: python

        history = [('Hello', 'Hi'), 'My Name is Victor', 'Nice to meet you']
        w = infer('Whats my name?', history_pairs=history_pairs).value()
        assert 'Victor' in w.value()

    Args:
        system_prompt (str or Text): Optional, the system prompt to use for completion.
        history_pairs (List[Tuple[str or Txt, str or Text]]): Optional, pairs of chat history that should preceed this axiomic.
        name (str): Optional, the name of the new axiomic.

    Returns:
        Text representing the completion.

    Raises:
        GraphError: If there was an error building this graph by the developer.
    '''

    return text.Text(prompt).infer(system_prompt=system_prompt, history_pairs=history_pairs, name=name)





import axiomic.providers.img_provider.together_img as together_img


def test_image_gen():
    img = together_img.gen_image("a white siamese cat", "stabilityai/stable-diffusion-xl-base-1.0", "1024x1024", "standard", 1)


if __name__ == '__main__':
    test_image_gen()




from typing import List, Dict

import axiomic

import axiomic.errors as errors
import axiomic.core.knits.json_pattern as json_pattern
import axiomic.core.knits.enforce_jsonschema as enforce_jsonschema
import json

import axiomic.data.lists as lists


def _make_jsonschema_enum(elements: List[str]) -> str:
    '''
    Make a JSON schema enum from a list of strings.

    Args:
        elements: A list of strings.

    Returns:
        A JSON schema enum.
    '''
    schema = {"type": "string", "enum": list(map(lambda s: s.lstrip('"').rstrip('"'), list(elements)))}
    return json.dumps(schema)


def _map_to_chat_history(case_map: Dict[str, List[str]]) -> List:
    l = []
    for k, v in case_map.items():
        for i in v:
            l.append((i, f'"{k}"'))
    return l


class Switch:
    '''
    Switch on an input and choose a specific case, effecively a semantic switch statement.
    For example, use this to choose which agent to route to based on the user input.

    .. code-block:: python

            import axiomic.builtins.knits as knits

            EXAMPLES = [
                ('I need help with the product I bought', '"SUPPORT"'),
                ('It is not working.', '"SUPPORT"'),
                ('I want to buy a product', '"SALES"'),
                ('I just bought a product and I need help.', '"SALES"'),
                ('I need a refund.', '"SALES"'),
                ('I do not like you. Go away.', '"COMPLAINT"')
            ]

            switch = knits.Switch(EXAMPLES)
            switch.infer('I just bought a product and I need help.').print()

    Args:
        case_examples: A dictionary mapping case names to lists of examples for that case.
        instructions: (Optional) explanation of the cases and when to choose which case.

    '''

    def __init__(self, case_examples: lists.ChatListType, instructions: axiomic.TextType = None):
        self.case_examples = lists.ChatList(case_examples)
        self.cases = [ex.agent for ex in self.case_examples]
        self.json_schema = _make_jsonschema_enum(list(set(self.cases)))
        self.json_gate = enforce_jsonschema.EnforceJsonSchema(schema=self.json_schema)
        if instructions is None:
            self.route_map_desc = 'Given text, state which of these cases most closely applies:\n'
            for k in self.cases:
                self.route_map_desc += f'{k}\n'
        else:
            self.route_map_desc = instructions
    
        for ex in self.case_examples:
            if not ex.agent.strip().startswith('"') or not ex.agent.strip().endswith('"'):
                raise errors.WeaveError(f'Case name must be a JSON serialized string (quoted): expected "{ex.agent}" got: {ex.agent}')
        self.json_pattern = json_pattern.JsonPattern(chat_list=self.case_examples, json_schema=self.json_schema, pattern_desc=self.route_map_desc)
        
    def infer(self, switch_text):
        '''
        :param case: The name of the agent to route to.
        :return: The name of the agent to route to.
        '''
        st = axiomic.Text(switch_text)
        return self.json_pattern.infer(switch_text)

    def make_eval_varients(self):
        case_vars = self.case_examples.make_eval_varients()
        return [(case_vars[0], Switch(case_vars[1])) for case_vars in case_vars]







import numpy as np

def cosine_similarity(vector1, vector2):
    dot_product = np.dot(vector1, vector2)
    norm_vector1 = np.linalg.norm(vector1)
    norm_vector2 = np.linalg.norm(vector2)
    similarity = dot_product / (norm_vector1 * norm_vector2)
    return similarity





import axiomic.logalytics as logalytics
import time

import axiomic.configure as configure


class RagProvider:
    '''
    Wrappr interface for all RAG implementations.
    '''

    def __init__(self, rag_impl):
        self.rag_impl = rag_impl
        configure.register_global_provider(self)

    def get_provider_name(self):
        return self.rag_impl.get_provider_name()

    def query_topk(self, query, k=3):

        with logalytics.RagQuery(self.get_provider_name(),
                             {'query': query,
                              'k': k,
                              'rag_provider': self.get_provider_name()}) as e:
            start = time.time()
            results = self.rag_impl.query(query, k)
            end = time.time()
            e.end(duration_s=end - start, k=k, results=results, rag_provider=self.get_provider_name())
        return results




import axiomic
import axiomic.builtins.knits.jsonschema_gate as jsonschema_gate
import json


class StringListSort:
    def __init__(self, criteria, 
                 instruction_template: str = '{p/builtin:knits:list_sort_template}', 
                 sysprompt_template: str = '{p/builtin:knits:list_sort_sysprompt}'):
        self.criteria = criteria
        self.check = jsonschema_gate.JsonSchemaGate('{p/builtin:jsonschema:list_of_string}')
        self.instruction_template = instruction_template
        self.tmpl = axiomic.Text(instruction_template)
        self.sysprompt = axiomic.Text(sysprompt_template)

    def infer(self, list_of_strings: list):
        serialized = json.dumps(list_of_strings)
        req = self.tmpl.format(criteria=self.criteria, list_of_strings=serialized)
        return self.check.infer(req.infer(system_prompt=self.sysprompt))
    
    def test_infer(self, w):
        pass
        





import axiomic.providers.llm_provider.llm_provider as llm_provider

import anthropic
import time
import os

import axiomic.errors as errors


if "ANTHROPIC_API_KEY" in os.environ:
    client = anthropic.Anthropic(
        # defaults to os.environ.get("ANTHROPIC_API_KEY")
        # api_key="my_api_key",
    )
else:
    client = None


# best_model = "claude-3-opus-20240229"
# model_sonnet = 'claude-3-sonnet-20240229'


def complete_history(prompt, model_name, max_tokens, system=None, history_pairs=[]):
    extra = {}
    if system:
        extra['system'] = system

    messages = [ ]

    for user, agent in history_pairs:
        messages.append({"role": "user", "content": user})
        messages.append({"role": "assistant", "content": agent})

    messages.append({"role": "user", "content": prompt})

    start = time.time()
    message = client.messages.create(
        model=model_name,
        max_tokens=max_tokens,
        messages=messages,
        **extra
    )
    end = time.time()
    
    return message.content[0].text, message.usage.input_tokens, message.usage.output_tokens, end - start


class AntropicLlmProvider:

    def get_default_context_params(self):
        return {'llm_provider_name': 'anthropic_text', 'llm_model_name': 'claude-3-sonnet-20240229', 'llm_temperature': 0.5, 'llm_max_tokens': 1024}

    def get_provider_name(self):
        return 'anthropic_text'

    def infer_history(self, llm_history_inference: llm_provider.LlmHistoryInferenceRequest) -> llm_provider.LlmInferenceResponse:
        retriable_error_text = None
        try:
            response, in_toks, out_toks, dur_s = complete_history(llm_history_inference.user_message, 
                                                                llm_history_inference.model_name, 
                                                                llm_history_inference.max_tokens, 
                                                                history_pairs=llm_history_inference.history_pairs,
                                                                system=llm_history_inference.system_prompt)
        except anthropic.InternalServerError as e:
            retriable_error_text = str(e)
        except anthropic.RateLimitError as e:
            retriable_error_text = str(e)

        if retriable_error_text:
            raise errors.ProviderErrorRetriable(retriable_error_text)

        resp = llm_provider.LlmInferenceResponse(
            llm_provider_name=llm_history_inference.llm_provider_name,
            model_name=llm_history_inference.model_name,
            temperature=llm_history_inference.temperature,
            max_tokens=llm_history_inference.max_tokens,
            response=response,
            input_tokens=in_toks,
            output_tokens=out_toks,
            duration_s=dur_s
        )
            
        return resp





import axiomic.core.knits as knits


def test_basic_switch():

    EXAMPLES = [
        ('I need help with the product I bought', '"SUPPORT"'),
        ('It is not working.', '"SUPPORT"'),
        ('I want to buy a product', '"SALES"'),
        ('I just bought a product and I need help.', '"SALES"'),
        ('I need a refund.', '"SALES"'),
        ('I do not like you. Go away.', '"COMPLAINT"')
    ]
    switch = knits.Switch(EXAMPLES)
    assert '"SUPPORT"' == switch.infer('I just bought a product and I need help.').value()




import hashlib
import base64

def hash_blake2(s):
    return hashlib.blake2b(s.encode()).hexdigest()

def hash_as_base32(s):
    hash_bytes = hashlib.blake2b(s.encode()).digest()
    return base64.b32encode(hash_bytes).decode('utf-8')


def short_hash(string):
    return hash_as_base32(string)[:12]




if __name__ == '__main__':
    print(short_hash('Hello, world!')) # 3


from openai import OpenAI

import axiomic.data.aimage as wimage

import axiomic.providers.img_provider.img_provider as img_provider
import time
import os


if "OPENAI_API_KEY" in os.environ:
    client = OpenAI()
else:
    client = None


def openai_image(prompt="a white siamese cat", model="dall-e-3", size_str="1024x1024", quality="standard", n=1):
    response = client.images.generate(
        model=model,
        prompt=prompt,
        size=size_str,
        quality=quality,
        n=n,
    )
    urls = [response.data[i].url for i in range(len(response.data))]
    imgs = []
    for url in urls:
        imgs.append(wimage.image_from_url(url))
    return imgs

# print(img)
# img.pretty_print()
# img.to_file('/tmp/woodchuck.png')

class OpenAiImageProviderImpl:
    def __init__(self):
        pass

    def get_default_context_params(self):
        return {
            'image_provider_name': 'openai_img',
            'image_model_name': 'dall-e-3',
            'image_width': 1024,
            'image_height': 1024,
        }
    
    def get_provider_name(self):
        return 'openai_img'

    def infer(self, req: img_provider.ImgInferenceRequest) -> img_provider.ImgInferenceResponse:
        size_str = str(req.image_width) + 'x' + str(req.image_height)
        start = time.time()
        images = openai_image(
            prompt=req.image_prompt,
            model=req.model_name,
            size_str=size_str,
            quality="standard",
            n=req.num_images
        )
        end = time.time()

        return img_provider.ImgInferenceResponse(
            req.img_provider_name,
            req.model_name,
            req.image_width,
            req.image_height,
            req.image_prompt,
            req.num_images,
            images,
            end - start
        )






import yaml
import axiomic
import axiomic.configure as configure

import axiomic.data.lists as lists

class Chat:
    '''
    A basic chat where the user says something and the agent responds.

    .. code-block:: python

        import axiomic.builtins.knits as knits
        chat = knits.Chat(system_prompt='Speak like a pirate')
        chat.infer('Hello! My favorite color is blue!').print()
        chat.infer('What is my favorite color?').print()

    By default, it will keep a conversation history. If you do not want a history to be
    kept, and want to do few shot prompting, you can set `append_mode=False`.

    This chat preseves the graph - it does not unaxiomic.

    Create a new chat.

    Args:
        chat_list: A list of chat pairs to start with. Default is no history.
        system_prompt: A prompt to use for how the agent should act over the conversation. Default is no specific system prompt.
        append_mode: If True, the chat will keep a history of the conversation from one call to `.infer()` to the next. If False, the only history will be `chat_pairs`.
    '''
    def __init__(self, chat_list: lists.ChatListType = None, 
                 system_prompt: axiomic.TextType = None, 
                 append_mode: bool = True):
        self.history = lists.ChatList(chat_list)
        self.system_prompt = system_prompt
        self.new_chats = lists.ChatList()
        self.append_mode = append_mode

    def get_history(self):
        '''
        Format the chat history as a string.
        '''
        return lists.ChatList(self.history.pairs + [(c.user.value(),  c.agent.value()) for c in self.new_chats.pairs])

    def set_system_prompt(self, system_prompt):
        '''
        Change the system prompt.
        
        Args:
            system_prompt: A prompt to use for how the agent should act over the conversation.
        '''
        self.system_prompt = system_prompt

    def set_append_mode(self, append_mode):
        '''
        Change if history is being kept.

        Args: 
            append_mode: If True, the chat will keep a history of the conversation from one call to `.infer()` to the next. If False, the only history will be `chat_pairs`.
        '''
        self.append_mode = append_mode

    def add_chat(self, user, agent):
        '''
        Inject a chat pair into the history as the most recent chat.

        Args:
            user: What the user said.
            agent: What the agent said.
        '''
        self.new_chats.append(user, agent)

    def add_chats(self, chat_list: lists.ChatListType):
        '''
        Add a list of chat pairs to the history as the the most recent chats.

        Args:
            chat_list: A list of chat pairs to add to the history.
        '''
        self.history.extend(chat_list)

    def infer(self, user, name=None):
        '''
        Send a user message to the the agent and get a response.

        If `append_mode` is True, the chat history will be updated with the new chat pair.

        Args:
            user: What the user said.
            name: A name to use for the Weave response. Default is unnamed (None).

        Returns:
            A weave which represents the agent's response.
        '''
        w = axiomic.Text(user)
        full_history = self.history.copy()
        full_history.extend(self.new_chats)
        resp = w.infer(system_prompt=self.system_prompt, history_pairs=full_history, name=name)
        if self.append_mode:
            self.add_chat(w, resp)
        return resp




from typing import List, Tuple

import axiomic.logalytics as logalytics

import dataclasses


@dataclasses.dataclass
class LlmHistoryInferenceRequest:
    llm_provider_name: str
    model_name: str
    temperature: float
    max_tokens: int
    user_message: str
    system_prompt: str
    history_pairs: List[Tuple[str, str]]


@dataclasses.dataclass
class LlmInferenceResponse:
    llm_provider_name: str
    model_name: str
    temperature: float
    max_tokens: int
    response: str
    input_tokens: int
    output_tokens: int
    duration_s: float


class LlmProvider:
    def __init__(self, llm_provider_ipml):
        self.llm_provider_ipml = llm_provider_ipml

    def get_default_context_params(self):
        return self.llm_provider_ipml.get_default_context_params()
    
    def get_provider_name(self):
        return self.llm_provider_ipml.get_provider_name()

    def infer_history(self, llm_history_inference: LlmHistoryInferenceRequest) -> LlmInferenceResponse:
        provider_name = llm_history_inference.llm_provider_name

        info = {
            'model_name': llm_history_inference.model_name,
            'temperature': llm_history_inference.temperature,
            'max_tokens': llm_history_inference.max_tokens,
            'user_message': llm_history_inference.user_message,
            'system_prompt': llm_history_inference.system_prompt,
            'history_pairs': llm_history_inference.history_pairs,
            'provider_name': provider_name
        }

        with logalytics.LlmInference(provider_name, info) as c:
            resp = self.llm_provider_ipml.infer_history(llm_history_inference)
            c.end(model_name=llm_history_inference.model_name, input_tokens=resp.input_tokens, output_tokens=resp.output_tokens, duration_s=resp.duration_s, response=resp.response, provider_name=provider_name)
            return resp




from typing import List, Tuple

import axiomic.logalytics as logalytics

import dataclasses


@dataclasses.dataclass
class ImgInferenceRequest:
    img_provider_name: str
    model_name: str
    image_width: int
    image_height: int
    image_prompt: str
    num_images: int


@dataclasses.dataclass
class ImgInferenceResponse:
    img_provider_name: str
    model_name: str
    image_width: int
    image_height: int
    image_prompt: str
    num_images: int
    image_list: List
    duration_s: float


class ImgProvider:
    def __init__(self, img_provider_impl):
        self.img_provider_impl = img_provider_impl

    def get_default_context_params(self):
        return self.img_provider_impl.get_default_context_params()
    
    def get_provider_name(self):
        return self.img_provider_impl.get_provider_name()

    def infer(self, req: ImgInferenceRequest) -> ImgInferenceResponse:
        provider_name = req.img_provider_name
        info = {
            'image_model_name': req.model_name,
            'image_width': req.image_width,
            'image_height': req.image_height,
            'image_prompt': req.image_prompt,
            'num_images': req.num_images,
            'provider_name': provider_name
        }

        with logalytics.ImageInference(provider_name, info) as c:
            resp = self.img_provider_impl.infer(req)
            c.end(image_list=resp.image_list, duration_s=resp.duration_s, image_model_name=resp.model_name,
                  image_width=resp.image_width, image_height=resp.image_height, provider_name=resp.img_provider_name)
        return resp
        #    c.end(model_name=llm_history_inference.model_name, input_tokens=resp.input_tokens, output_tokens=resp.output_tokens, duration_s=resp.duration_s, response=resp.response)
        #    return resp






import os
import inspect


def second_caller_filename():
    # Get the current stack trace
    stack = inspect.stack()
    
    if len(stack) < 3:
        return None

    # Get the second stack frame
    frame = stack[2]
    
    # Get the filename of the second stack frame
    filename = frame.filename
    
    # Return the filename
    return filename



    




import axiomic.utils.naming as naming
import axiomic.errors as errors

Name = naming.Name

def test_full_names():
    assert naming.is_full_name('Bucket:Dir:Base')
    assert naming.is_full_name('Bucket-_012:D02-_ir:B12-_ase')
    assert not naming.is_full_name('_Bucket-_012:D02-_ir:B12-_ase')
    assert not naming.is_full_name('-Bucket-_012:D02-_ir:B12-_ase')
    assert not naming.is_full_name('0Bucket-_012:D02-_ir:B12-_ase')
    assert not naming.is_full_name(' 0Bucket-_012:D02-_ir:B12-_ase')
    assert not naming.is_full_name('Bucket:D ir:Base')
    assert not naming.is_full_name('Bucket:D@ir:Base')


def test_refernce_types():
    assert Name('p/Bucket:Dir:Base').type_name == 'p'
    assert Name('var/Bucket:Dir:Base').type_name == 'var'
    assert Name('pair/Bucket:Dir:Base').type_name == 'pair'
    assert Name('pair/Base').type_name == 'pair'



def test_get_embedded_refs():
    story = '''
Hello there. {p/foo} is here. and {p/buck:di:name} wow! 
And {p/uh:oh} is also there.
    '''
    hits = naming.get_imbedded_refs(story)
    assert hits == ['p/foo', 'p/buck:di:name', 'p/uh:oh']





import inspect
import json
import os
import rich
import rich.panel as rich_panel
import yaml

import axiomic.configure as configure
import axiomic.engine as engine
import axiomic.engine.functional as nF
import axiomic.errors as errors
import axiomic.protos as protos
import axiomic.utils.naming as naming

import axiomic.providers.embedding_provider.embedding_provider as embedding_provider

import axiomic.frontend.expand as expand
import axiomic.frontend.image as image
import axiomic.frontend.embedding as embedding

import axiomic.models as models

import axiomic.utils.wjson as wjson
import typing

import axiomic.data.lists as lists


DEFAULT_CONTEXT = None

class VarContext:
    def __init__(self, contents):
        self.name = contents['name']
        self.default = contents['default']


class ParamContext:
    def __init__(self, contents):
        self.name = contents['name']
        self.default = contents['default']


def handle_missing_param_file(param_name, filepath):
    raise errors.GraphError(f'Param file for \'{param_name}\' not found: {filepath}')


class TextContext:
    def __init__(self):
        self.base_dir = './weave_context'
        self._vars = {}
        self._params = {}

    def var_filename(self, var_name):
        return f'{self.base_dir}/{var_name}.var.yaml'

    def param_filename(self, param_name):
        p = f'{self.base_dir}/{param_name}.param.yaml'
        return os.path.abspath(p)

    def _load_var(self, var_name):
        with open(self.var_filename(var_name), 'r') as f:
            v = yaml.safe_load(f)
        self._vars[var_name] = VarContext(v)

    def load_param(self, param_name):
        if not os.path.exists(self.param_filename(param_name)):
            handle_missing_param_file(param_name, self.param_filename(param_name))

        with open(self.param_filename(param_name), 'r') as f:
            v = yaml.safe_load(f)
        self._params[param_name] = ParamContext(v)
        
    def var(self, var_name):
        v = self._vars.get(var_name)
        return Text(v.default, name=var_name)

    def param(self, param_name):
        if param_name not in self._params:
            self.load_param(param_name)

        p = self._params.get(param_name)
        return Text(p.default, name=param_name)



def _get_caller_info(levels_up=1):
    """
    Returns the filename and line number of the caller up N levels in the call stack.
    
    Args:
    levels_up (int): Number of levels up in the stack from the current function to find the caller.
    
    Returns:
    tuple: A tuple containing the filename and line number of the caller.
    """
    try:
        frame = inspect.currentframe()
        for _ in range(levels_up + 1):  # +1 to skip the current function's frame
            frame = frame.f_back
            if frame is None:
                return None, None  # No caller at this level
        return frame.f_code.co_filename, frame.f_lineno
    finally:
        del frame  # Prevent reference cycles


# Prints a nice string
def pretty_print_box(title, contents, levels_up=1, nav_breadcrumbs={}):
    for nav_key in nav_breadcrumbs:
        rich.print(f'[grey50]{nav_breadcrumbs[nav_key]} ({nav_key})[/grey50]')
    panel = rich_panel.Panel(contents, title=title, expand=False, title_align="left", border_style="grey50", style="grey50")
    rich.print(panel)


def default_context():
    global DEFAULT_CONTEXT
    if DEFAULT_CONTEXT is None:
        DEFAULT_CONTEXT = TextContext()
    return DEFAULT_CONTEXT


def var(name: str):
    return default_context().var(name)


def param(name: str):
    return default_context().param(name)


def recursive_imbed_refs(text: str) -> str:
    imbed_refs = naming.get_imbedded_refs(text)

    if len(imbed_refs) == 0:
        return text

    ref_provider = configure.infer_context().provider_provider.ref_store_provider
    for ref in imbed_refs:
        exact_substr = '{' + ref + '}'
        ref_value = ref_provider.get_default_value(ref)

        ref_value_recur = recursive_imbed_refs(ref_value)
        text = text.replace(exact_substr, ref_value_recur)
    
    return text

def ref(ref_name):
    name = naming.Name(ref_name)
    ref_provider = configure.infer_context().provider_provider.ref_store_provider
    ref_value = ref_provider.get_default_value(ref_name)
    nav_breadcrumbs = (ref_provider.get_ref_nav_breadcrumbs(ref_name))
    return Text(ref_value, name=name.original_name, nav_breadcrumbs=nav_breadcrumbs)


class Text:
    def __init__(self, node: any, eager=False, json=False, name=None, nav_breadcrumbs={}, autoexpand=True):
        self.eager = eager
        self.nav_breadcrumbs = nav_breadcrumbs
        self._value = None
        self.graph_builder = nF.GraphBuilder()
        self.given_name = name
        if isinstance(node, Text):
            self._clone_from(node)
        elif isinstance(node, protos.axiomic.AxiomicNode):
            self.node = node
            self.graph_builder.add_node(self.node)
            if name is not None:
                if name != node.name:
                    raise NotImplementedError('Name passing not supported yet')
        elif not isinstance(node, str):
            raise errors.GraphError(f'Must be a string: {node}')
        elif autoexpand and expand.text_needs_expand(node):
            expanded_into_weave = expand.recur_build_wave_with_subs(node, Text)
            self._clone_from(expanded_into_weave)
        else:
            self.node = nF.weave(self.graph_builder, node, name=name)
            self.graph_builder.add_node(self.node)
        expand.track_text(self)

    def __format__(self, format_spec):
        return expand.text_format_place_holder(self)

    def _clone_from(self, w):
        self.node = w.node
        self.eager = w.eager
        # TODO should this be a deep copy?
        self.graph_builder = w.graph_builder

    def _merger(self):
        m = Merger(eager=self.eager)
        m.merge_from(self)
        return m

    def infer_image(self, n=1, name=None):
        m = self._merger()
        c = models.infer_context()
        params = protos.axiomic.TextToImagesParams(
            image_provider_name=c.image_provider_name, 
            model_name=c.image_model_name, 
            image_width=c.image_width, 
            image_height=c.image_height, 
            num_images=n
        )
        img_infer_node = nF.infer_image(self.graph_builder, 
                                      self.node, 
                                      params=params,
                                      name=name)
        wi = image.Image(img_infer_node, eager=self.eager, name=name)
        m.merge_into(wi)
        return wi

    def print(self, verbose=True):
        '''
        Prints the resolved value of this Text to the console.
        This is not considered to value as it does not return the value to you.

        It will force resolution of this weave, and any associated LLM inferences which haven't bene completed yet.
        Therefore, it may raise any errors associated with `value()`, including a ThoughtCrime.
        
        Raises:
            ThoughtCrime: If there was an error building this weave by the AI.
            GraphError: If there was an error building this weave by the developer.
        '''
        name = self.given_name
        if name is None:
            name = ''
        else:
            name = f' - {name}'

        nav_breadcrumbs = {}
        if verbose:
            nav_breadcrumbs = dict(self.nav_breadcrumbs)
            filename, line_number = _get_caller_info(1)
            nav_breadcrumbs['printed'] = f'{filename}:{line_number}'
        
        pretty_print_box(f'Text{name}', self._unsafe_access(), nav_breadcrumbs=nav_breadcrumbs)
        return self

    def infer(self, system_prompt=None, history_pairs: lists.ChatListType = None, name=None):
        '''
        Does LLM text compleition on this weave, returning a Text which represents the completed text.

        The model used for completion is determined by the context, for example,

        .. code-block:: python

            import axiomic.quick_config as qc
            with qc.OpenAiGpt4:
                w = axiomic.Text('What is the capital of France?')
                assert w.infer().value() == 'Paris'


        .. code-block:: python

            history = [('Hello', 'Hi'), ('My Name is Victor', 'Nice to meet you')]
            w = axiomic.Text('Whats my name?')
            assert 'Victor' in w.infer(history_pairs=history_pairs).value()

        Args:
            system_prompt (str or Text): Optional, the system prompt to use for completion.
            history_pairs (List[Tuple[str or Text, str or Weave]]): Optional, pairs of chat history that should preceed this axiomic.
            name (str): Optional, the name of the new axiomic.

        Returns:
            A Text representing the completion.

        Raises:
            GraphError: If there was an error building this weave by the developer.
        '''
        history_pairs = lists.ChatList(history_pairs)

        m = self._merger()
        if system_prompt is None:
            system_prompt = m.autoweave(Text(''))
        else:
            system_prompt = m.autoweave(system_prompt)
        history_pair_nodes = []

        for pair in history_pairs:
            user = m.autoweave(pair.user)
            agent = m.autoweave(pair.agent)
            history_pair_nodes.append((user.node, agent.node))

        c = models.infer_context()
        complete_params = protos.axiomic.CompleteParams(
            llm_provider_name=c.llm_provider_name, llm_model_name=c.llm_model_name, 
            llm_temperature=c.llm_temperature, llm_max_tokens=c.llm_max_tokens,
        )
        return m.subweave(nF.complete(self.graph_builder, 
                                      self.node, 
                                      system_prompt.node, 
                                      history_pairs=history_pair_nodes, 
                                      complete_params=complete_params,
                                      name=name))

    def format(self, **inputs):
        '''
        Preforms string formatting on this axiomic. Either strings or weaves can be formatted.
        
        .. code-block:: python

            color = axiomic.Text('Blue')
            template = axiomic.Text('The sky is {sky_color} and the grass is {grass_color}')
            result = template.format(sky_color=color, grass_color='Green')
            assert result.value() == 'The sky is Blue and the grass is Green'

        Args:
            **inputs: Mapping of name to str or Text to format into this axiomic.

        Returns:
            A new Text that contains the values formatted together.

        Raises:
            GraphError: If formatting incorrectly.
        '''
        m = self._merger()
        boxed_inputs = {}
        for k, v in inputs.items():
            boxed_inputs[k] = m.autoweave(v).node
        return m.subweave(nF.format(self.graph_builder, self.node, **boxed_inputs))
    
    def name(self, name: str):
        m = self._merger()
        return m.subweave(nF.name_node(self.graph_builder, value=self.node, name=name))

    def _run(self):
        self._value = engine.eager_weft(self.graph_builder.get_graph(), self.node)

    def __str__(self):
        # TODO determine if we should allow str to implicitly value
        return self.__format__(None)
        # return str(self.value())

    def __repr__(self):
        return f'Text({self.node.name})'

    def value(self):
        '''
        Returns the string value of this axiomic. This may result in multple calls to
        multiple LLMs to resolve the value. Note: all previous calls were deferred
        until unaxiomic.
        
        .. code-block:: python

            w = axiomic.Text('What is the capital of France?')
            assert w.value() == 'What is the capital of France?'
            assert w.infer().value() == 'Paris'

        Returns:
            The string value of this axiomic.

        Raises:
            ThoughtCrime: If there was an error building this weave by the AI.
            GraphError: If there was an error building this weave by the developer.
        '''
        self._run()
        return self._value

    def value_json(self):
        '''
        values as a JSON, returning the value. Raises a GraphError if the value is not a valid JSON.
        You should consider using .validate_json(schema) instead if you want to ensure this is
        a valid JSON before unweaving it.
        
        .. code-block:: python

            w = axiomic.Text('{"name": "Alice"}')
            assert w.value_json()['name'] == 'Alice'

        Returns:
            The JSON value (dict, list, str, int, float, bool, or None) of this axiomic.

        Raises:
            GraphError: If the schema is not a valid JSON schema.
        '''
        return wjson.loads(self.value())

    def check(self, check):
        '''
        Checks this weave against the given check, returning either "true" or "false".

        Args:
            check: A check from the `axiomic.core.checks` module.

        Returns:
            A new Text (boolean, either "true" or "false") which indicates if this weave passes the check.
        '''
        m = self._merger()
        if check.is_core_unary_check():
            code = check.get_core_unary_check_code()
            return m.subweave(nF.core_unary_check(self.graph_builder, self.node, code))
        elif check.is_core_binary_check():
            code = check.get_core_binary_check_code()
            input_secondary = m.autoweave(check.get_input_secondary())
            return m.subweave(nF.builtin_binary_check(self.graph_builder, self.node, input_secondary.node, code))
        else:
            raise errors.GraphError('Not given a core unary or binary check')

    def checkpoint(self, check):
        '''
        Checkpoints this weave on the given check, causing a thought crime if the check fails.

        If the check succeeds, the weave returns itself.

        Args:
            check: A check from the `axiomic.core.checks` module.

        Returns:
            This axiomic.

        Raises:
            ThoughtCrime: If the check fails.
        '''
        m = self._merger()
        condition = m.autoweave(self.check(check))
        return m.subweave(nF.boolean_gate(self.graph_builder, self.node, condition.node))

    def validate_json(self, schema):
        '''
        Validates if this weave is a JSON which matches the given schema.

        Args:
            schema (str or Text): The schema to validate against.

        Returns:
            A new Text (boolean, either "true" or "false") which indicates if this weave follows the given JSON schema.

        Raises:
            GraphError: If the schema is not a valid JSON schema.
        '''
        m = self._merger()
        schema = m.autoweave(schema)
        return m.subweave(nF.check_json_schema(self.graph_builder, self.node, schema.node))

    def print_graph(self):
        '''
        Prints the execution graph of this weave to the console.
        '''
        engine.print_graph(self.graph_builder.get_graph(), self.node)

    def _rag_topk(self, rag_provider, k=3, separator='\n'):
        m = self._merger()
        rag_provider_name = rag_provider.get_provider_name()
        return m.subweave(nF.flat_rag_topk_query(self.graph_builder, self.node, separator=separator, k=k, rag_provider=rag_provider_name))

    def embed(self):
        c = models.infer_context()
        return embedding.TextEmbedding(self.node, c.embedding_provider_name, c.embedding_model_name, eager=self.eager)

    def _unsafe_access(self):
        '''Intenral use only - accesses the value without going through an unaxiomic.
        '''
        self._run()
        return self._value


TextType = typing.Union[str, Text]


class Merger:
    def __init__(self, eager=True):
        self.eager = eager
        self.weaves = []

    def __call__(self, w: Text):
        self.merge_from(w)
        return w
    
    def merge_from(self, w: Text):
        self.weaves.append(w)
    
    def merge_into(self, w: Text):
        for m in self.weaves:
            w.graph_builder.merge_from(m.graph_builder)
        return w

    def autoweave(self, thing: any):
        if is_json_type(thing):
            w = Text(json.dumps(thing), eager=self.eager)
            self.merge_from(w)
            return w
        if isinstance(thing, Text):
            self.merge_from(thing)
            return thing
        if thing is None:
            raise errors.GraphError('Cannot weave None')
        w = Text(thing, eager=self.eager)
        self.merge_from(w)
        return w

    def subweave(self, node):
        w = Text(node, eager=self.eager)

        self.merge_into(w)
        return w


JSON_TYPES = set([
    dict,
    list,
    # str, - doesn't count :)
    int,
    float,
    bool,
])

def is_json_type(thing: any):
    return type(thing) in JSON_TYPES
    





import axiomic.errors as errors
import axiomic.engine.functional as nF
import axiomic.protos as protos
import axiomic.engine as engine
import axiomic.data.aimage as wimage


class Image:
    '''
    An Image. Do not call the constructor.
    '''
    def __init__(self, node: any, eager=False, name=None):
        self.eager = eager
        # self.nav_breadcrumbs = nav_breadcrumbs
        self._value = None
        self.graph_builder = nF.GraphBuilder()
        self.given_name = name
        #if isinstance(node, Weave):
        #    raise errors.WeaveError('Image Weaves cannot be created from Weave objects.')
        if isinstance(node, protos.axiomic.AxiomicNode):
            self.node = node
            self.graph_builder.add_node(self.node)
            if name is not None:
                if name != node.name:
                    raise NotImplementedError('Name passing not supported yet')
        else:
            raise errors.WeaveError('Image must be constructed from a weave graph node.')

    def infer(self):
        '''
        Not Implemented Yet. Image to text coming soon.
        '''
        raise NotImplementedError('Image Weaves do not support infer yet: Infer on images not implemented yet')

    def _run(self):
        if self._value is None:
            self._value = engine.eager_weft(self.graph_builder.get_graph(), self.node)

    def __str__(self):
        return self.value()

    def __repr__(self):
        return f'Image({self.node.name})'

    def value(self):
        '''
        Returns:
            The image as a WImage.
        '''
        self._run()
        return self.value_image()

    def print_graph(self):
        engine.print_graph(self.graph_builder.get_graph(), self.node)

    def value_image(self, index=0):
        '''
        Returns:
            The image as a WImage.
        '''
        self._run()
        return self._value[index]

    def print(self):
        '''
        Prints the image to the terminal.
        '''
        self._run()
        for img in self._value:
            img.print()
        return self






import axiomic.models as models

import axiomic.errors as errors

import axiomic.utils.paramtool as paramtool
import weakref
import re





'''
This implements recursive parameter expansion through subsitution of a string.

For example:
"My {P.foo}" -> "My {P.whiz} and {P.bang}" -> "My cat and dog"

This is done through creating texts for each component, and then building a tree which is
seralized into the string.

"My {P.foo}" -> "My {textref-12345}"
and textref-12345 =  "{weaveref-22345} and {weaveref-32345}"
textref-22345 = "cat
textref-32345 = "dog"
'''

ACTIVE_textS = weakref.WeakValueDictionary()
text_REF_REGEX = re.compile(r'{(weaveref_0x[a-z0-9]+)}')


def magic_name(w):
    return 'textref_' + hex(id(w))


def list_text_magic_names(text: str) -> list:
    return text_REF_REGEX.findall(text)


def get_param_by_ref(param_ref: str) -> str:
    ref_provider = models.infer_context().provider_provider.ref_store_provider
    return ref_provider.get_default_value(param_ref)


def track_text(w):
    ACTIVE_textS[magic_name(w)] = w


def text_format_place_holder(w):
    return f'{{{magic_name(w)}}}'


def swap_param_for_text(text: str, param_name, w) -> str:
    text_magic_name = magic_name(w)
    p_format_name = f'{{{param_name}}}'
    return text.replace(p_format_name, f'{{{text_magic_name}}}')

def get_active_text(weave_magic_name: str):
    if text_magic_name not in ACTIVE_WEAVES:
        raise errors.textError(f'(Internal Error - please report) No such active weave: {weave_magic_name}')
    return ACTIVE_textS[weave_magic_name]


def text_needs_expand(text: str) -> bool:
    params_mentioned = paramtool.get_all_imbedded_params(text)
    if params_mentioned:
        return True

    texts_mentioned = list_text_magic_names(text)
    if texts_mentioned:
        return True

    return False


def recur_build_wave_with_subs(text: str, Text_: callable):
    '''
    This will build out a text graph while subsitute param values.
    '''
    params_mentioned = paramtool.get_all_imbedded_params(text)
    formats = {}
    for param_name in params_mentioned:
        param_text = get_param_by_ref(param_name)
        param_as_text = recur_build_wave_with_subs(param_text, Text_)
        text = swap_param_for_text(text, param_name, param_as_text)
        formats[magic_name(param_as_text)] = param_as_text

    for text_magic_name in list_text_magic_names(text):
        w = get_active_text(weave_magic_name)
        formats[text_magic_name] = w

    if not formats:
        w = Text_(text)
        track_text(w)
        return w


    # We need to find text refs as well
    w = Text_(text, autoexpand=False)
    track_text(w)
    w = w.format(**formats)
    track_text(w)
    return w




import axiomic
import os
import argparse
import rich
import dataclasses
from typing import List


@dataclasses.dataclass
class ParamListTable:
    param_ref: List[str]
    file_path: List[str]
    value_preview: List[str]


def print_param_list_table(table: ParamListTable):
    table_rows = []
    for param_ref, file_path, value_preview in zip(table.param_ref, table.file_path, table.value_preview):
        table_rows.append([param_ref, file_path, value_preview])
    table_console = rich.table.Table(show_header=True, header_style="bold blue")
    table_console.add_column("Param Ref")
    table_console.add_column("File Path")
    table_console.add_column("Value Preview")
    for row in table_rows:
        table_console.add_row(*row)
    console = rich.console.Console()
    console.print(table_console)


def _recur_dir(dirname):
    filenames = []

    if os.path.isdir(dirname):
        for filename in os.listdir(dirname):
            filepath = os.path.join(dirname, filename)
            if os.path.isdir(filepath):
                sub_names = _recur_dir(filepath)
                filenames.extend(sub_names)
            else:
                filenames.append(filepath)

    return filenames


def path_to_param_ref(p):
    if not p.startswith('p/'):
        raise ValueError(f"Invalid param path: {p}")
    part = p[2:].replace('/', '.')
    if part.endswith('.txt'):
        part = part[:-4]
    return f'P.{part}'


def get_param_preview(data_root, param_path):
    full_path = os.path.join(data_root, param_path)
    with open(full_path, 'r') as file:
        content = file.read().replace('\n', ' ')
    return content[:60]


def list_param_files(data_rot):
    filenames = _recur_dir(os.path.join(data_rot, 'p'))
    # strip data_rot prefix
    filenames = [filename[len(data_rot):] for filename in filenames]
    return filenames


def build_param_table(data_rot):
    param_file_list = list_param_files(data_rot)

    table = ParamListTable([], [], [])

    for param_path in param_file_list:
        table.param_ref.append(path_to_param_ref(param_path))
        table.file_path.append(os.path.join(data_rot, param_path))
        table.value_preview.append(get_param_preview(data_rot, param_path))
    return table


def handle_create_param(data_rot, name):
    if not name.startswith('P.'):
        raise ValueError(f"Invalid param name: {name}")
    filepath = os.path.join(data_rot, name.replace('.', '/') + '.txt')

    dirname = os.path.dirname(filepath)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    if not os.path.exists(filepath):
        with open(filepath, 'w') as file:
            file.write("")
    
    print(f'Created file: {filepath}')


def main():
    parser = argparse.ArgumentParser(description="Axiomic Data Tool")
    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand")

    # Create the parser for the "list" subcommand
    list_parser = subparsers.add_parser("list", help="List parameters in Axiomic data root")
    list_parser.add_argument("--data_rot", help="Path to Weave data root")

    # Create the parser for the "create" subcommand
    create_parser = subparsers.add_parser("create", help="Create a new parameter")
    create_parser.add_argument("--data_rot", help="Path to Axiomic data root")
    create_parser.add_argument("name", help="Name of the parameter")

    args = parser.parse_args()

    if args.subcommand == "list":
        param_table = build_param_table(args.data_rot)
        print_param_list_table(param_table)
    elif args.subcommand == "create":
        handle_create_param(args.data_rot, args.name)
    else:
        print("Invalid subcommand")

if __name__ == "__main__":
    main()
    



import axiomic.protos as protos

from google.protobuf import descriptor
import json


from rich.console import Console
from rich.table import Table
from rich.tree import Tree
from rich.panel import Panel
from rich import box
import dataclasses

class Graph:
    def __init__(self, graph: protos.axiomic.AxiomicGraph):
        self.graph = graph
        self._nodes = {}

        for node in graph.nodes:
            self._nodes[node.name] = node

    def get_node(self, name: str) -> protos.axiomic.AxiomicNode:
        return self._nodes[name]

    def has_node(self, name: str) -> bool:
        return name in self._nodes

    def ref_to_node(self, ref: protos.axiomic.NodeRef) -> protos.axiomic.AxiomicNode:
        return self.get_node(ref.name)

    def flatten_node(self, node: protos.axiomic.AxiomicNode) -> list:
        return reversed_breadth_first_flatten(self, node)

    def print_tree(self, node: protos.axiomic.AxiomicNode):
        _print_tree(self, node)


def get_subnode(node: protos.axiomic.AxiomicNode) -> protos.axiomic.AxiomicNode:
    field_name = node.WhichOneof('node_type')
    sub_node = getattr(node, field_name)
    return field_name, sub_node


def is_axiomic_node(thing):
    return str(type(thing)) == str(protos.axiomic.NodeRef)


def list_format_node_childern_extras(graph: Graph, fn: protos.axiomic.FormatNode) -> list:
    l = []
    for format_input in fn.inputs:
        child = graph.ref_to_node(format_input.value)
        l.append((f'{{{format_input.name}}}', child))
    return l

def list_history_complete_node_childern_extras(graph: Graph, hn: protos.axiomic.HistoryCompleteNode) -> list:
    l = []
    for i, history_pair in enumerate(hn.chat_history):
        l.append((f'user_message_{i}', graph.ref_to_node(history_pair.user_message)))
        l.append((f'assistant_message_{i}', graph.ref_to_node(history_pair.assistant_message)))
    return l

SPECIAL_CASE_LIST = {
    'format_node': list_format_node_childern_extras,
    'history_complete_node': list_history_complete_node_childern_extras,
}

def list_childern(graph: Graph, node: protos.axiomic.AxiomicNode) -> list:
    sub_node_name, sub_node = get_subnode(node)
    descriptor = sub_node.DESCRIPTOR
    childern = []
    for field in descriptor.fields:
        field_name = field.name
        if is_axiomic_node(getattr(sub_node, field_name)):
            childern.append((field_name, graph.ref_to_node(getattr(sub_node, field_name))))
    
    if sub_node_name in SPECIAL_CASE_LIST:
        childern += SPECIAL_CASE_LIST[sub_node_name](graph, sub_node)
    return childern


def is_autogenerated_name(name: str) -> bool:
    return name.startswith('_')


def list_childern_named_only(graph: Graph, node: protos.axiomic.AxiomicNode) -> list:
    childern = list_childern(graph, node)

    # if a child has an autogenerated name, recurse to get another name for it.
    for i, (child_name, child_node) in enumerate(childern):
        ...


def str_parent(parent_name, childern):
    childern_str = ''
    for child in childern:
        childern_str += f'  {child[0]}: {child[1].name}'
    childern_str = childern_str.strip()
    return (parent_name, childern_str)


def abbrivate_str(s: str, max_len=80) -> str:
    if len(s) == 0:
        return '(EMPTY STRING)'
    s = s.strip().replace('\n', ' ').strip()
    if len(s) > max_len:
        return s[:max_len] + '...'
    return s


def is_leaf(graph: Graph, node: protos.axiomic.AxiomicNode) -> bool:
    return len(list_childern(graph, node)) == 0


def str_leaf(node: protos.axiomic.AxiomicNode) -> str:
    field_name = node.WhichOneof('node_type')

    # print(node)
    if field_name == 'immediate_node':
        abbrivated_val = abbrivate_str(node.immediate_node.value)
        name = node.name
        return name, abbrivated_val
    return name, 'LEAF'


def dedup_childern_list(childern: list) -> list:
    childern_set = set()
    deduped_childern = []
    for child in childern:
        if child.name not in childern_set:
            deduped_childern.append(child)
            childern_set.add(child.name)
    return deduped_childern


def _reversed_breadth_first_flatten(graph: Graph, node: protos.axiomic.AxiomicNode) -> list:
    childern = list_childern(graph, node)
    ans = []
    for child_name, child_node in childern:
        ans.extend(_reversed_breadth_first_flatten(graph, child_node))
    ans.append(node)
    return ans


def _print_node(graph: Graph, node, childern):
    if is_leaf(graph, node):
        return str_leaf(node)
    else:
        return str_parent(node.name, childern)


def reversed_breadth_first_flatten(graph: Graph, node: protos.axiomic.AxiomicNode) -> list:
    childern = _reversed_breadth_first_flatten(graph, node)
    deduped_childern = dedup_childern_list(childern)
    return deduped_childern


def _print_tree(graph: Graph, node: protos.axiomic.AxiomicNode):
    print_nodes = build_print_nodes(graph, node)

    for print_node in print_nodes:
        print_tree_node(print_node)


def extract_props_from_json(j: dict) -> dict:
    '''
    Convert from a json object to a list of properties. 
    Recusive pull out any string key/value pairs which are not names.
    Names are handled as childern nodes.
    '''
    props = {}
    for k, v in j.items():
        if k == 'name':
            continue
        if isinstance(v, dict):
            props.update(extract_props_from_json(v))
        if not isinstance(v, dict) and not isinstance(v, list):
            props[k] = str(v)
    return props


def get_node_props(node: protos.axiomic.AxiomicNode) -> dict:
    props = extract_props_from_json(json.loads(protos.json_dump(node)))
    # Detect empty immediate nodes and insert a special empty string place holder
    if node.WhichOneof('node_type') == 'immediate_node' and 'value' not in props:
        props['value'] = '[underline]<Empty String>[/underline]'
    return props


def __print_tree(graph: Graph, node: protos.axiomic.AxiomicNode):
    nodes = reversed_breadth_first_flatten(graph, node)

    console = Console()
    table = Table(show_header=False, header_style="dim", box=None, title=f"Graph: {node.name}", title_style="dim")

    table.add_column("Node Name", style="dim", justify="right")
    table.add_column("Node Value", style="dim")
    for node in nodes:
        childern = list_childern(graph, node)
        name, stuff = _print_node(graph, node, childern)
        table.add_row(name, stuff)
    console.print("\n")
    console.print(table)
    console.print("\n")


def node_to_print_node(graph: Graph, node: protos.axiomic.AxiomicNode):
    print_name = node.name
    props = get_node_props(node)
    childern = []

    for child_name, child_node in list_childern(graph, node):
        child_print_node = node_to_print_node(graph, child_node)
        childern.append(TreeChild(child_name, child_print_node))

    return TreePrintNode(print_name, props, childern)


def build_print_nodes(graph: Graph, node: protos.axiomic.AxiomicNode):
    nodes = reversed_breadth_first_flatten(graph, node)

    print_nodes = []
    for node in nodes:
        # childern = list_childern(graph, node)
        # name, stuff = _print_node(graph, node, childern)
        print_nodes.append(node_to_print_node(graph, node))

    return print_nodes


@dataclasses.dataclass
class TreeChild:
    relation: str
    child_node: any


@dataclasses.dataclass
class TreePrintNode:
    name: str
    props: dict
    children: list[TreeChild]

    
def node_short_desc(child_node: TreeChild):
    if len(child_node.child_node.children) == 0:
        if 'value' in child_node.child_node.props:
            panel = Panel(child_node.child_node.props["value"], 
                          title=f'{child_node.relation}: {child_node.child_node.name}',
                          title_align='left',
                          expand=False)
            return panel
            # return f'{child_node.relation}: {child_node.child_node.name} - {child_node.child_node.props["value"]}'
    return f'{child_node.relation}: {child_node.child_node.name}'


def recur_add_childern(child_node, wip_tree, max_depth=4):

    for gchild in child_node.children:
        gchild_relation = gchild.relation
        gchild_name = gchild.child_node.name
        gchild_tree = wip_tree.add(node_short_desc(gchild))
        
        if max_depth > 0:
            recur_add_childern(gchild.child_node, gchild_tree, max_depth - 1)
        else:
            gchild_tree.add('...')


def print_tree_node(node: TreePrintNode):
    console = Console()
    #console.print(f"[bold]{node.name}[/bold]")

    if len(node.props) > 0:
        props_table = Table(show_header=False, header_style="", show_lines=True, show_edge=True,
                            title=node.name, expand=False)
        props_table.add_column("Property")
        props_table.add_column("Value")
        for k, v in sorted(node.props.items()):
            props_table.add_row(str(k), str(v))
        # console.print(props_table)
        tree = Tree(props_table)
    else:
        tree = Tree(node.name)
    # tree.add(props_table)
    for child in node.children:
        ctree = tree.add(node_short_desc(child))
        recur_add_childern(child.child_node, ctree, max_depth=2)
    console.print(tree)


if __name__ == '__main__':

    node = TreePrintNode('root', {'a': 1, 'b': 2}, [TreeChild('input', 'child1'), TreeChild('output', 'child2')])
    print_tree_node(node)
    



import axiomic.configure as configure

import axiomic.core.knits.chat as chat
import axiomic.core.checks as checks

import axiomic.data.lists as lists

import yaml
import axiomic


class JsonPattern:
    '''
    Have the agent continue a pattern genering a JSON response (i.e. few shot prompting).

    .. code-block:: python

        import axiomic.builtins.knits as knits
        examples = [('Jack and Jill', '[{"name": "Jack"}, {"name": "Jill"}]')]   
        examples.append(('Jake out ran Mark", [{"name": "Jake"}, {"name": "Mark"}]'))
        json_schema = '{ "type": "array", "items": { "type": "object", "properties": { "name": { "type": "string" } }, "required": ["name"] } }'
        pattern = knits.JsonPattern(examples, json_schema)

        resp = pattern.infer('The 1967 film Bonnie and Clyde is directed by Arthur Penn')
        assert resp.unweave_json() == [{"name": "Bonnie"}, {"name": "Clyde"}, {"name": "Arthur Penn"}]

    Args:
        chat_list: A few examples of user-agent interactions to extrapolate from.
        json_schema: A JSON schema that the response must match.
        pattern_desc: (optional) A description of the pattern.
        system_prompt_template: (optional) Defaults to `'{P.builtin.knits.json_pattern_system_prompt}'`. Only recommended to set this if the default isn't working or you want to save tokens.
        append_mode: Defaults to false, but will keep adding each new response as a new few shot example, eventually becoming many shot.

    This pattern preseves the graph - it does not unaxiomic.
    '''
    def __init__(self, chat_list: lists.ChatListType, 
                 json_schema: axiomic.TextType, 
                 pattern_desc: axiomic.TextType = 'Continue the pattern', 
                 system_prompt_template: axiomic.TextType = None,
                 append_mode=False):
        self.chat_list = lists.ChatList(chat_list)
        self.pattern_desc = pattern_desc
        self.append_mode = append_mode
        if system_prompt_template is None:
            system_prompt_template = '{P.core.knits.json_pattern.system_prompt}'
        self.system_prompt_template = system_prompt_template
        self.system_prompt = axiomic.format(system_prompt_template, pattern_desc=pattern_desc, schema=json_schema)
        self.json_schema = json_schema
        self.json_schema_checked = axiomic.Text(json_schema).checkpoint(checks.CheckJsonSyntax())
        self.check_json_schema = checks.CheckJsonSchema(self.json_schema)
        self.chat = chat.Chat(chat_list=self.chat_list, system_prompt=self.system_prompt, append_mode=append_mode)

    def infer(self, prompt, name=None):
        '''
        Infers the next response in the pattern.

        Args:
            prompt: The user prompt.

        Returns:
            The agent response as a Weave, checked against the JSON schema.
        '''
        w = axiomic.Text(prompt)
        resp = self.chat.infer(w, name=name).checkpoint(checks.CheckJsonSyntax()).checkpoint(self.check_json_schema)
        return resp






from typing import Any

import axiomic.errors as errors

class ProviderProvider:
    def __init__(self):
        self.providers = {}
        self.ref_store_provider = None

    def register_provider(self,  provider):
        self.providers[provider.get_provider_name()] = provider

    def get_provider(self, name) -> Any:
        if name not in self.providers:
            raise errors.WeaveError(f'Provider {name} not found.')
        return self.providers[name]
    
    def get_ref_store_provider(self):
        return self.ref_store_provider
    




class StringList:
    def __init__(self, strings=[]):
        self.strings = strings

    def extend(self, strings):
        self.strings.extend(strings)
    
    def append(self, string):
        self.strings.append(string)



from typing import List, Tuple
import yaml

import axiomic.utils.paramtool as paramtool
import axiomic.errors as errors



class ParamProvider:
    '''
    A reference store holds parameters, templates, etc. and information about them.
    '''
    def __init__(self, impl):
        self.impl = impl

    def exists(self, name) -> bool:
        if not paramtool.is_param_name(name):
            return False
        return self.impl.exists(name)

    def is_bucket(self, name) -> bool:
        if not paramtool.is_param_name(name):
            return False
        return self.impl.is_bucket(name)
    
    def get_nav_breadcrumbs(self, name) -> str:
        return self.impl.get_ref_nav_breadcrumbs(name)

    def get_default_value(self, name) -> str:
        if not paramtool.is_param_name(name):
            raise errors.WeaveError(f'Invalid param name: {name}')
        return self.impl.get_default_value(name)

    def get_pairs(self, name) -> List[Tuple[str, str]]:
        raw_text = self.get_default_value(name)
        a_b_pairs = yaml.safe_load(raw_text)
        l = []
        for d in a_b_pairs:
            l.append((d['A'].strip(), d['B'].strip()))
        return l




import axiomic.core_config as core_config


class ToDeleteConfig:
    def __init__(self, cfg):
        self.cfg = cfg

    def _write(self):
        core_config.set_global_config(self.cfg)

    def disable_file_cache(self):
        self.cfg.enable_file_cache = False
        self._write()

    def enable_file_cache(self):
        self.cfg.enable_file_cache = True
        self._write()

    def disable_caching(self):
        self.cfg.enable_caching = False
        self._write()

    def enable_caching(self):
        self.cfg.enable_caching = True
        self._write()

    def set_thought_crime_retry_count(self, count):
        self.cfg.thought_crime_retry_count = count
        self._write()


# config = WeaveConfig(core_config.global_config)



import axiomic.models as configure

import axiomic.logalytics.explainers as explainers

from axiomic.frontend.text import Text
import axiomic.errors as errors


# Maps param names to text
PARAM_MAP = { }


def is_bucket(param_ref: str) -> bool:
    ref_provider = configure.infer_context().provider_provider.ref_store_provider
    return ref_provider.is_bucket(param_ref)

def is_base_param(param_ref: str) -> bool:
    ref_provider = configure.infer_context().provider_provider.ref_store_provider
    return ref_provider.exists(param_ref)

def get_param_by_ref(param_ref: str) -> str:
    if param_ref not in PARAM_MAP:
        ref_provider = configure.infer_context().provider_provider.ref_store_provider
        w = Text(ref_provider.get_default_value(param_ref))
        PARAM_MAP[param_ref] = w
    return PARAM_MAP[param_ref]

def get_filename_by_ref(param_ref: str) -> str:
    ref_provider = configure.infer_context().provider_provider.ref_store_provider
    return list(ref_provider.get_nav_breadcrumbs(param_ref).values())[0]


class ParamBucket:
    def __init__(self, name_parts: list):
        '''
        List of strings that represent the name of the parameter.
        For example, 'P.foo.bar' would be ['foo', 'bar']
        '''
        self.name_parts = name_parts

    def __getattr__(self, name):
        new_name = self.__as_ref(name)
        if is_bucket(new_name):
            return ParamBucket(self.name_parts + [name])
        if is_base_param(new_name):
            return get_param_by_ref(new_name)
        want_file = get_filename_by_ref(new_name)
        explainers.missing_file_ref(new_name, want_file)

    def __as_ref(self, subname: str = None):
        if subname is None:
            return '.'.join(self.name_parts)
        return '.'.join(self.name_parts + [subname])
        
P = ParamBucket(['P'])




import axiomic.providers.rag_provider.rag_provider as rag_provider
import axiomic.providers.rag_provider.demo_rag as demo_rag

RagProvider = rag_provider.RagProvider
DemoRagProviderImpl = demo_rag.DemoRagProviderImpl




import axiomic.providers.llm_provider.llm_provider as llm_provider

from together import Together
import time
import os

API_KEY = os.environ.get("TOGETHER_API_KEY", None)

if API_KEY:
    client = Together(api_key=os.environ.get("TOGETHER_API_KEY"))
else:
    client = None


def complete_history(prompt, model_name, max_tokens, system=None, history_pairs=[]):
    messages = [ ]

    if system:
        messages.append({"role": "system", "content": system})

    for user, agent in history_pairs:
        messages.append({"role": "user", "content": user})
        messages.append({"role": "assistant", "content": agent})

    messages.append({"role": "user", "content": prompt})

    start = time.time()

    response = client.chat.completions.create(
        model=model_name,
        messages=messages,
        max_tokens=max_tokens,
    )

    end = time.time()
    return response.choices[0].message.content, response.usage.prompt_tokens, response.usage.completion_tokens, end - start


class TogetherProviderLlmImpl:

    def get_default_context_params(self):
        return {'llm_provider_name': 'together_text', 'llm_model_name': 'meta-llama/Llama-3-8b-chat-hf', 'llm_temperature': 0.5, 'llm_max_tokens': 1024}

    def get_provider_name(self):
        return 'together_text'

    def infer_history(self, llm_history_inference: llm_provider.LlmHistoryInferenceRequest) -> llm_provider.LlmInferenceResponse:
        response, in_toks, out_toks, dur_s = complete_history(llm_history_inference.user_message, 
                                                            llm_history_inference.model_name, 
                                                            llm_history_inference.max_tokens, 
                                                            history_pairs=llm_history_inference.history_pairs,
                                                            system=llm_history_inference.system_prompt)

        resp = llm_provider.LlmInferenceResponse(
            llm_provider_name=llm_history_inference.llm_provider_name,
            model_name=llm_history_inference.model_name,
            temperature=llm_history_inference.temperature,
            max_tokens=llm_history_inference.max_tokens,
            response=response,
            input_tokens=in_toks,
            output_tokens=out_toks,
            duration_s=dur_s
        )
            
        return resp
    






import pydantic

import axiomic.core.uagents.pydantic_pattern as pydantic_pattern


class COTTest(pydantic.BaseModel):
    rationale: str
    answer: bool


EXAMPLES = [
    ('a car is bigger than an apple.', COTTest(rationale='In general, cars weight more than apples', answer=True)),
    ('A tree is smaller than a pencile', COTTest(rationale='in general, penciles are made of part of a tree, so a tree cannot be smaller', answer=False)),
]


def test_basic():
    pat = pydantic_pattern.PydanticPattern(EXAMPLES, 'Explain your thinking then evalute the statements as true or false')
    assert pat.infer('dogs are bigger than cats').answer == True
    assert pat.infer('a mountain is weighs less than a person').answer == False




import re

import axiomic.logalytics.explainers as explainers


IMPLIED_BUCKET = 'default'
IMPLIED_DIRNAME = 'default'

NAME_BASENAME_PAT = r'[a-zA-Z][a-zA-Z0-9_.-]+'
NAME_DIRNAME_PAT  = r'[a-zA-Z][a-zA-Z0-9_-]+'
NAME_BUCKET_PAT = r'[a-zA-Z][a-zA-Z0-9_-]*'

# Matches <BUCKET>:<SUBBUCKET>:<BASENAME>
FULL_NAME_REGEX = re.compile(f'^(?P<bucket>{NAME_BUCKET_PAT}):(?P<dir>{NAME_DIRNAME_PAT}):(?P<basename>{NAME_BASENAME_PAT})$')
NAME_IMPLIED_BUCKET_REGEX = re.compile(f'^(?P<dir>{NAME_DIRNAME_PAT}):(?P<basename>{NAME_BASENAME_PAT})$')
NAME_IMPLIED_DIR_REGEX = re.compile(f'^(?P<basename>{NAME_BASENAME_PAT})$')


REFERENCE_TYPES = ['p', 'var', 'pair']
REFERENCE_TYPES_OR_PAT = '|'.join(REFERENCE_TYPES)

# For example, match "param/Bucket:Dir:Base"
# FULL_REFERENCE_REGEX = re.compile(f'^(?P<type>{REFERENCE_TYPES_OR_PAT})/...
FULL_REFERENCE_REGEX = re.compile(f'^(?P<type>{REFERENCE_TYPES_OR_PAT})/(?P<bucket>{NAME_BUCKET_PAT}):(?P<dir>{NAME_DIRNAME_PAT}):(?P<basename>{NAME_BASENAME_PAT})$')
REFERENCE_IMPLIED_BUCKET_REGEX = re.compile(f'^(?P<type>{REFERENCE_TYPES_OR_PAT})/(?P<dir>{NAME_DIRNAME_PAT}):(?P<basename>{NAME_BASENAME_PAT})$')
REFERFENCE_IMPLIED_DIR_REGEX = re.compile(f'^(?P<type>{REFERENCE_TYPES_OR_PAT})/(?P<basename>{NAME_BASENAME_PAT})$')


IMMBEDDED_REF_PAT = f'{{(?P<type>{REFERENCE_TYPES_OR_PAT})/(?P<bucket>{NAME_BUCKET_PAT}:)?(?P<dir>{NAME_DIRNAME_PAT}:)?(?P<basename>{NAME_BASENAME_PAT})}}'
IMBEDDED_REF_REGEX = re.compile(IMMBEDDED_REF_PAT)


def get_imbedded_refs(text: str) -> list:
    hits = IMBEDDED_REF_REGEX.findall(text)

    l = []
    for h in hits:
        r = h[0] + '/'
        stuff = ''.join(h[1:])
        l.append(r + stuff)
    return l


def is_name(name: str) -> bool:
    if FULL_NAME_REGEX.match(name) is not None:
        return True
    
    if NAME_IMPLIED_BUCKET_REGEX.match(name) is not None:
        return True
    
    if NAME_IMPLIED_DIR_REGEX.match(name) is not None:
        return True

    return False


def convert_to_full_name(name: str) -> str:
    if m := FULL_NAME_REGEX.match(name):
        return name
    
    if m := NAME_IMPLIED_BUCKET_REGEX.match(name):
        return f'{IMPLIED_BUCKET}:{m.group("dir")}:{m.group("basename")}'

    if m := NAME_IMPLIED_DIR_REGEX.match(name):
        return f'{IMPLIED_BUCKET}:{IMPLIED_DIRNAME}:{m.group("basename")}'


def extract_name_parts(name: str) -> tuple:
    if m := FULL_NAME_REGEX.match(name):
        return m.group('bucket'), m.group('dir'), m.group('basename')
    return None, None, None


def is_full_name(name: str) -> bool:
    return FULL_NAME_REGEX.match(name) is not None


def is_full_reference_name(name: str) -> bool:
    return FULL_REFERENCE_REGEX.match(name) is not None


def is_reference_name(name: str) -> bool:
    if is_full_reference_name(name):
        return True

    if REFERENCE_IMPLIED_BUCKET_REGEX.match(name):
        return True
    
    if REFERFENCE_IMPLIED_DIR_REGEX.match(name):
        return True
    
    return False
    

def convert_to_full_reference_name(name: str) -> str:
    if m := FULL_REFERENCE_REGEX.match(name):
        return name
    
    if m := REFERENCE_IMPLIED_BUCKET_REGEX.match(name):
        return f'{m.group("type")}/{IMPLIED_BUCKET}:{m.group("dir")}:{m.group("basename")}'
    
    if m := REFERFENCE_IMPLIED_DIR_REGEX.match(name):
        return f'{m.group("type")}/{IMPLIED_BUCKET}:{IMPLIED_DIRNAME}:{m.group("basename")}'


def extract_reference_parts(name: str) -> tuple:
    if m := FULL_REFERENCE_REGEX.match(name):
        return m.group('type'), m.group('bucket'), m.group('dir'), m.group('basename')
    return None, None, None, None


class Name:
    def __init__(self, name: str):
        if isinstance(name, Name):
            self._copy(name)
            return

        if name.startswith('{') and name.endswith('}'):
            explainers.explain_wrapped_name(name)

        if name is None:
            explainers.explain_bad_name_error('<None>')

        self.original_name = name
        self.type_name = None
        self.bucket_name = None
        self.dir_name = None
        self.base_name = None
        self.is_reference = False

        if is_reference_name(name):
            self.full_name = convert_to_full_reference_name(name)
            self.type_name, self.bucket_name, self.dir_name, self.base_name = extract_reference_parts(self.full_name)
            self.is_reference = True
        elif is_name(name):
            self.full_name = convert_to_full_name(name)
            self.bucket_name, self.dir_name, self.base_name = extract_name_parts(self.full_name)
        else:
            explainers.explain_bad_name_error(name)

    def _copy(self, other):
        self.original_name = other.original_name
        self.type_name = other.type_name
        self.bucket_name = other.bucket_name
        self.dir_name = other.dir_name
        self.base_name = other.base_name
        self.is_reference = other.is_reference

    def __str__(self):
        return self.full_name


        






import axiomic.configure as configure

import axiomic.core.knits.chat as chat

import axiomic.core.knits.json_pattern as json_pattern 

import yaml
import axiomic

import axiomic.data.lists as lists
import json
import pydantic


def pydantic_to_str(pydantic_obj):
    if not isinstance(pydantic_obj, pydantic.BaseModel):
        return pydantic_obj
    
    return pydantic_obj.json()


class PydanticPattern:
    '''
    Continues a given pattern which produces Pydnatic objects (i.e. few shot prompting).

    .. code-block:: python

        import axiomic.builtins.uagents as uagents
        import pydantic
                
        class COTAnswer(pydantic.BaseModel):
            rationale: str
            answer: bool

        EXAMPLES = [
            ('a car is bigger than an apple.', COTAnswer(rationale='In general, cars weight more than apples', answer=True)),
            ('A tree is smaller than a pencile', COTAnswer(rationale='in general, penciles are made of part of a tree, so a tree cannot be smaller', answer=False)),
        ]

        pattern = pydantic_pattern.PydanticPattern(EXAMPLES, 'Explain your thinking then evalute the statements as true or false')
        assert pattern.infer('dogs are bigger than cats').answer == True

    Args:
        examples: A few examples of user-agent interactions to extrapolate from. The user message must be a string. The agent message must be pydnatic object.
        pattern_desc: A description of the pattern.
        system_prompt_template: (optional) Defaults to a reasonable template. Only recommended to set this if the default isn't working or you want to save tokens.

    This is a micro agent - it will `unweave()` when creating the response.
    '''
    def __init__(self, examples: lists.ChatListType, pattern_desc: axiomic.TextType, system_prompt_template: axiomic.TextType = None):
        self.chat_list = lists.ChatList(examples)
        self.pattern_desc = pattern_desc
        if len(examples) == 0:
            raise ValueError('examples must be non-empty')
        self.examples = examples
        self.json_schema = json.dumps(examples[0][1].model_json_schema())
        self.serialized_examples = lists.ChatList([(pydantic_to_str(e[0]), e[1].json()) for e in examples])
        self.json_pattern = json_pattern.JsonPattern(chat_list=self.serialized_examples, 
                                                     json_schema=self.json_schema, 
                                                     pattern_desc=pattern_desc, 
                                                     system_prompt_template=system_prompt_template)
        self.ModelClass = self.examples[0][1].__class__

    def infer(self, prompt, name=None):
        '''
        Returns the pydantic object that matches the pattern for `prompt`.

        Since this returns a pydnatic object, it will `unweave()` the response and the graph won't be propogated.

        Args:
            prompt: The next prompt in the pattern.

        Returns:
            The agent response as a Pydantic object.
        '''
        resp = self.json_pattern.infer(pydantic_to_str(prompt), name=name)
        j = resp.value_json()
        return self.ModelClass(**j)




from typing import List, Tuple

import axiomic.logalytics as logalytics

import dataclasses


@dataclasses.dataclass
class EmbeddingRequest:
    embedding_provider_name: str
    embedding_model_name: str
    text: str


@dataclasses.dataclass
class EmbeddingResponse:
    embedding_provider_name: str
    embedding_model_name: str
    embedding: list
    duration_s: float


class EmbeddingProvider:
    def __init__(self, impl):
        self.impl = impl

    def get_default_context_params(self):
        return self.impl.get_default_context_params()
    
    def get_provider_name(self):
        return self.impl.get_provider_name()

    def infer(self, req: EmbeddingRequest) -> EmbeddingRequest:
        # provider_name = llm_history_inference.llm_provider_name

        resp = self.impl.infer(req)
        return resp

        # info = {
        #     'model_name': llm_history_inference.model_name,
        #     'temperature': llm_history_inference.temperature,
        #     'max_tokens': llm_history_inference.max_tokens,
        #     'user_message': llm_history_inference.user_message,
        #     'system_prompt': llm_history_inference.system_prompt,
        #     'history_pairs': llm_history_inference.history_pairs
        # }

        #with logalytics.LlmInference(provider_name, info) as c:
        #    resp = self.llm_provider_ipml.infer_history(llm_history_inference)
        #    c.end(model_name=llm_history_inference.model_name, input_tokens=resp.input_tokens, output_tokens=resp.output_tokens, duration_s=resp.duration_s, response=resp.response)
        #    return resp




import os
from together import Together
import axiomic.data.aimage as wimage
import axiomic.providers.img_provider.img_provider as img_provider
import time

if "TOGETHER_API_KEY" in os.environ:
    client = Together(api_key=os.environ.get("TOGETHER_API_KEY"))
else:
    client = None

def gen_image(prompt="a white siamese cat", model="dall-e-3", size_str="1024x1024", quality="standard", n=1):
    response = client.images.generate(
        prompt=prompt,
        model=model,
        steps=10,
        n=1,
    )
    return [wimage.image_from_base64(response.data[0].b64_json)]


class TogetherImageProviderImpl:
    def __init__(self):
        pass

    def get_default_context_params(self):
        return {
            'image_provider_name': 'together_img',
            'image_model_name': 'stabilityai/stable-diffusion-2-1', # 'stabilityai/stable-diffusion-xl-base-1.0',
            'image_width': 1024,
            'image_height': 1024,
        }
    
    def get_provider_name(self):
        return 'together_img'

    def infer(self, req: img_provider.ImgInferenceRequest) -> img_provider.ImgInferenceResponse:
        size_str = str(req.image_width) + 'x' + str(req.image_height)
        start = time.time()
        images = gen_image(
            prompt=req.image_prompt,
            model=req.model_name,
            size_str=size_str,
            quality="standard",
            n=req.num_images
        )
        end = time.time()

        return img_provider.ImgInferenceResponse(
            req.img_provider_name,
            req.model_name,
            req.image_width,
            req.image_height,
            req.image_prompt,
            req.num_images,
            images,
            end - start
        )







# import axiomic.builtins.knits.jsonschema_gate as jsonschema_gate
import axiomic.core.knits.pattern as pattern
import axiomic.core.knits.chat as chat
import axiomic.core.knits.json_pattern as json_pattern
import axiomic.core.knits.switch as switch

# JsonSchemaGate = jsonschema_gate.JsonSchemaGate
Pattern = pattern.Pattern
Chat = chat.Chat
JsonPattern = json_pattern.JsonPattern
Switch = switch.Switch




import os

import axiomic.configure.default_config as default_config
import axiomic.errors as errors
import axiomic.logalytics.explainers as explainers
import axiomic.utils.paramtool as paramtool

WEAVEROOT = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
BUILTIN_ROOT = os.path.join(WEAVEROOT, 'core_data')


def handle_missing_param_file(param_name, filepath):
    raise errors.WeaveError(f'Param file for \'{param_name}\' not found: {filepath}')


class FilesystemParamProviderImpl:
    '''
    A filesystem backed reference store.
    '''
    def __init__(self):
        self._root_dir = None

    def exists(self, name) -> bool:
        return os.path.exists(self._filename(name))

    def is_bucket(self, name) -> bool:
        '''
        :param name: the name of the param or bucket

        :return: True if the name is a bucket, and the bucket exists.
        '''
        filename = self._filename(name, suffix='')
        return os.path.isdir(filename)

    def _get_root_dir(self):
        if self._root_dir is None:
            self._root_dir = default_config.file_ref_store.data_path
        if self._root_dir is None:
            explainers.set_file_store_path()
        return os.path.abspath(self._root_dir)

    def get_ref_nav_breadcrumbs(self, name) -> str:
        filename = self._filename(name, '.txt')
        return {'file ref': filename}

    def get_default_value(self, name) -> str:
        '''
        Returns the default value for a given name by reading the store.
        '''
        return self._read_file(name)

    def _filename(self, name, suffix: str = '.txt') -> str:
        parts = paramtool.segment_param_name(name)
        if parts[0] == 'core':
            return self._builtin_filename(name, suffix)
        return self._external_filename(name, suffix)

    def _builtin_filename(self, name, suffix: str) -> str:
        parts = paramtool.segment_param_name(name)
        builtin_path = os.path.join(BUILTIN_ROOT, 'P', *parts[:-1], f'{parts[-1]}{suffix}')
        return builtin_path

    def _external_filename(self, name, suffix: str) -> str:
        root = self._get_root_dir()
        parts = paramtool.segment_param_name(name)
        return os.path.join(root, 'P', *parts[:-1],  f'{parts[-1]}{suffix}')

    def _assert_file_exists(self, name, suffix: str = '.txt'):
        filepath = self._filename(name, suffix)
        if not os.path.exists(filepath):
            explainers.missing_file_ref(name, filepath)

    def _read_file(self, name) -> str:
        self._assert_file_exists(name)
        with open(self._filename(name), 'r') as f:
            return f.read()






import math
from collections import Counter

import re

import axiomic.providers.rag_provider.rag_provider as rag_provider



def split_doc(doc):
    words = re.sub(r'[^a-z0-9 \n]', '', doc.lower())
    parts = words.split()
    return parts



    

    

class TfidfVectorizer:
    def __init__(self):
        self.vocab = {}
        self.idf = {}

    def fit(self, documents):
        unique_words = set()
        for doc in documents:
            words = split_doc(doc)
            unique_words.update(words)

        self.vocab = {word: idx for idx, word in enumerate(unique_words)}

        N = len(documents)
        split_docs = [split_doc(doc) for doc in documents]
        for word in self.vocab:
            # df = sum(1 for doc in documents if word in split_doc(doc))
            df = sum(1 for doc in split_docs if word in doc)
            self.idf[word] = math.log(N / df)

    def transform(self, documents):
        vectors = []
        for doc in documents:
            words = split_doc(doc)
            word_counts = Counter(words)
            doc_len = len(words)

            vector = [0] * len(self.vocab)
            for word, count in word_counts.items():
                if word in self.vocab:
                    tf = count / doc_len
                    idx = self.vocab[word]
                    vector[idx] = tf * self.idf[word]

            vectors.append(vector)

        return vectors


def cosine_similarity(vec1, vec2):
    dot_product = sum(a * b for a, b in zip(vec1, vec2))
    magnitude1 = math.sqrt(sum(a ** 2 for a in vec1))
    magnitude2 = math.sqrt(sum(b ** 2 for b in vec2))
    if magnitude1 == 0 or magnitude2 == 0:
        return 0
    else:
        return dot_product / (magnitude1 * magnitude2)


def get_nns_by_vector(query_vector, vectors, k):
    similarities = []
    for i, vector in enumerate(vectors):
        similarity = cosine_similarity(query_vector, vector)
        similarities.append((i, similarity))

    similarities.sort(key=lambda x: x[1], reverse=True)
    nearest_neighbors = [x[0] for x in similarities[:k]]
    return nearest_neighbors
    

class DemoRagProviderImpl:
    '''
    This is not suitable for production use. This is intended for Experiments, Demos and POCs only.

    This is a Demo implementation of a RAG Store.  
    It offers poor quality and slow performance, but it's very simple and portable.
    It is a top-k query across a corpus of text using TF-IDF vectorization and cosine similarity. 

    '''

    def __init__(self):
        self.vectorizer = TfidfVectorizer()
        self.texts = []
        self.vectors = []

        # Register Globally
        rag_provider.RagProvider(self)

    def get_provider_name(self):
        return f'Temporary#DemoRagProviderImpl-{id(self)}'

    def fit_texts(self, fit_texts):
        self.texts.extend(fit_texts)
        self.vectorizer.fit(self.texts)
        self.vectors = self.vectorizer.transform(self.texts)

    def query(self, query, k=3):
        query_words = split_doc(query)
        query_vector = [0] * len(self.vectorizer.vocab)

        for word in query_words:
            if word in self.vectorizer.vocab:
                idx = self.vectorizer.vocab[word]
                query_vector[idx] = self.vectorizer.idf[word]

        similar_indices = get_nns_by_vector(query_vector, self.vectors, k)
        return [self.texts[idx] for idx in similar_indices]


if __name__ == '__main__':
    # Sample usage
    texts = [
        "The quick brown fox jumps over the lazy dog",
        "The lazy dog sleeps all day long",
        "The quick brown fox is very quick and jumps high",
        "The dog is lazy and sleeps a lot",
        "The fox is brown and quick in jumping",
        "The cat is orange and prowls all day long",
        "The cat is very active and jumps high",
    ]

    rag = DemoRagProviderImpl()
    rag.fit_texts(texts)

    print(rag.query("orange cat jumps", k=3))





import time
from openai import OpenAI

import axiomic.providers.embedding_provider.embedding_provider as embedding_provider

import os


if "OPENAI_API_KEY" in os.environ:
    client = OpenAI()
else:
    client = None


def embed(text, model_name):
    start = time.time()
    message = client.embeddings.create(
        input=text,
        model=model_name
    )
    end = time.time()
    return message.data[0].embedding, end - start


class OpenAiLlmProviderImpl:
    def get_default_context_params(self):
        default_model = 'text-embedding-3-small' # 'text-embedding-ada-002' #  # 
        return {'embedding_provider_name': 'openai_embed', 'embedding_model_name': default_model}

    def get_provider_name(self):
        return 'openai_embed'

    def infer(self, req: embedding_provider.EmbeddingRequest) -> embedding_provider.EmbeddingRequest:
        embedding, dur_s = embed(req.text, req.embedding_model_name)

        resp = embedding_provider.EmbeddingResponse(
            embedding_provider_name=req.embedding_provider_name,
            embedding_model_name=req.embedding_model_name,
            embedding=embedding,
            duration_s=dur_s
        )
        return resp


if __name__ == '__main__':
    print(embed('Hello, world!', 'text-embedding-ada-002'))



import axiomic.configure as configure

import axiomic.core.knits.chat as chat
import axiomic.data.lists as lists

import yaml
import axiomic



class Pattern:
    def __init__(self, chat_list, system_prompt=None):
        self.chat_list = lists.ChatList(chat_list)
        self.chat_list.print()
        self.chat = chat.Chat(chat_list=self.chat_list, system_prompt=system_prompt, append_mode=False)
        

    def infer(self, next_a, name=None):
        w = axiomic.Text(next_a)
        return self.chat.infer(w, name=name)




import axiomic

import axiomic.protos as protos


def _unary_check(cls):
    def is_core_unary_check(self) -> bool:
        return True
    cls.is_core_unary_check = is_core_unary_check
    def is_core_binary_check(self) -> bool:
        return False
    cls.is_core_binary_check = is_core_binary_check
    def get_core_unary_check_code(self) -> int:
        return self.check_code
    cls.get_core_unary_check_code = get_core_unary_check_code
    def get_core_binary_check_code(self) -> int:
        raise ValueError('This is not a binary check.')
    cls.get_core_binary_check_code = get_core_binary_check_code
    def get_input_secondary(self) -> axiomic.TextType:
        raise ValueError('This is not a binary check.')
    cls.get_input_secondary = get_input_secondary
    return cls
    

def _binary_check(cls):
    def is_core_unary_check(self) -> bool:
        return False
    cls.is_core_unary_check = is_core_unary_check
    def is_core_binary_check(self) -> bool:
        return True
    cls.is_core_binary_check = is_core_binary_check
    def get_core_unary_check_code(self) -> int:
        raise ValueError('This is not a unary check.')
    def get_core_binary_check_code(self) -> int:
        return self.check_code
    cls.get_core_binary_check_code = get_core_binary_check_code
    def get_input_secondary(self) -> axiomic.TextType:
        return self.input_secondary 
    cls.get_input_secondary = get_input_secondary
    return cls


@_unary_check
class CheckJsonSyntax:
    def __init__(self):
        self.check_code = protos.axiomic.BuiltinUnaryCheckNode.CHECK_JSON_SYNTAX
    
    def __call__(self, possibly_json: axiomic.TextType) -> axiomic.Text:
        '''
        Check if the input is valid JSON.

        Args:
            possibly_json: A string that may be valid JSON.

        Returns:
            Text which represents if the input is valid JSON.
        '''
        return axiomic.Text(possibly_json).check(self)


@_binary_check
class CheckJsonSchema:
    def __init__(self, schema: axiomic.TextType):
        self.input_secondary = schema
        self.check_code = protos.axiomic.BuiltinBinaryCheckNode.CHECK_JSONSCHEMA

    def __call__(self, json_object: axiomic.TextType) -> axiomic.Text:
        '''
        Check if the input is valid JSON.

        Args:
            json_object: A string that may be valid JSON.
            schema: A string that may be valid JSON schema.

        Returns:
            Text which represents if the input is valid JSON.
        '''
        return axiomic.Text(json_object).check(self)
    



import axiomic.providers.llm_provider.llm_provider as llm_provider
import axiomic.providers.llm_provider.anthropic_llm as anthropic_llm
import axiomic.providers.llm_provider.openai_llm as openai_llm
import axiomic.providers.param_provider.filesys_params as fs_params
import axiomic.providers.param_provider.param_provider as param_provider
import axiomic.providers.img_provider.img_provider as img_provider
import axiomic.providers.img_provider.openai_img as openai_img
import axiomic.providers.embedding_provider.openai_embed as openai_embed
import axiomic.providers.embedding_provider.embedding_provider as embedding_provider
import axiomic.providers.llm_provider.together_llm as together_llm
import axiomic.providers.img_provider.together_img as together_img


FILESYSTME_PARAMS = param_provider.ParamProvider(fs_params.FilesystemParamProviderImpl())
TOGETHER_LLM = llm_provider.LlmProvider(together_llm.TogetherProviderLlmImpl())
TOGETHER_IMG = img_provider.ImgProvider(together_img.TogetherImageProviderImpl())
ANTHROPIC_LLM = llm_provider.LlmProvider(anthropic_llm.AntropicLlmProvider())
OPENAI_LLM = llm_provider.LlmProvider(openai_llm.OpenAiLlmProvider())
OPENAI_IMG = img_provider.ImgProvider(openai_img.OpenAiImageProviderImpl())
OPENAI_EMBED = embedding_provider.EmbeddingProvider(openai_embed.OpenAiLlmProviderImpl())





import axiomic
import axiomic.core.checks as checks

class EnforceJsonSchema:
    def __init__(self, schema):
        self.schema = schema
        self.checked_schema = axiomic.Text(self.schema).check(checks.CheckJsonSyntax())

    def infer(self, json_str):
        w = axiomic.Text(json_str)
        return w.checkpoint(checks.CheckJsonSyntax()).check(self.checked_schema)
    



import re


PARAM_BASE_PATTERN = r'P(?:\.[a-zA-Z0-9_]+)+'
PARAM_BASE_REGEX = re.compile(f'^{PARAM_BASE_PATTERN}$')
PARAM_IN_BRACKETS_REGX = re.compile(f'{{(?P<param_str>{PARAM_BASE_PATTERN})}}')


def is_param_name(thing: str) -> bool:
    return PARAM_BASE_REGEX.match(thing) is not None


def segment_param_name(param_name: str) -> list:
    return param_name.split('.')[1:]

def get_all_imbedded_params(text: str) -> list:
    if type(text) is not str:
        raise TypeError(f"Expected a string, got {type(text)}: {text}")
    hits = PARAM_IN_BRACKETS_REGX.findall(text)
    return [h for h in hits]




import axiomic.utils.paramtool as paramtool 
import axiomic.errors as errors

def test_param_names():
    assert paramtool.is_param_name('P.foo')
    assert paramtool.is_param_name('P.foo.bar')
    assert not paramtool.is_param_name('P.foo.bar.')
    assert not paramtool.is_param_name('P.foo.-bar')
    assert not paramtool.is_param_name('P.foo bar')
    assert not paramtool.is_param_name('P.')
    assert not paramtool.is_param_name('P')


def test_segment_name():
    assert paramtool.segment_param_name('P.foo') == ['foo']
    assert paramtool.segment_param_name('P.foo.bar') == ['foo', 'bar']
    assert paramtool.segment_param_name('P.foo.bar.Whi8z') == ['foo', 'bar', 'Whi8z']


def test_get_imbedded_params():
    assert paramtool.get_all_imbedded_params('{P.foo}') == ['P.foo']
    assert paramtool.get_all_imbedded_params('Hello {P.foo}!') == ['P.foo']
    assert paramtool.get_all_imbedded_params('My {P.foo} is {P.bar}') == ['P.foo', 'P.bar']



